from abc import abstractmethod, ABC
import os
import numpy as np
import casadi as ca
import forwardkinematics.urdfFks.casadiConversion.urdfparser as u2c

from forwardkinematics.fksCommon.fk import ForwardKinematics

class URDFForwardKinematics(ForwardKinematics):

    def __init__(self, fileName, links, rootLink, n):
        super().__init__()
        self._urdf_file = os.path.dirname(os.path.abspath(__file__)) + "/urdf/" + fileName
        self._links = links
        self._rootLink = rootLink
        self.readURDF()
        self._n = n
        self._q_ca = ca.SX.sym('q', n)
        self.generateFunctions()

    def readURDF(self):
        self.robot = u2c.URDFparser()
        self.robot.from_file(self._urdf_file)

    def generateFunctions(self):
        fks = []
        for i in range(self._n + 1):
            fks.append(self.casadi(self._q_ca, i))
        self._fks_fun = ca.Function('fk', [self._q_ca], fks)

    def casadi(self, q: ca.SX, i, positionOnly=False):
        tip = self._links[i]
        if positionOnly:
            return self.robot.get_forward_kinematics(self._rootLink, tip, q)['T_fk'][0:3, 3]
        else:
            return self.robot.get_forward_kinematics(self._rootLink, tip, q)['T_fk']

    def numpy(self, q: np.ndarray, i, positionOnly=False):
        if positionOnly:
            return np.array(self._fks_fun(q)[i])[0:3, 3]
        else:
            return np.array(self._fks_fun(q)[i])



"""Functions for getting casadi expressions for quaternions from joint type."""
import casadi as cs
import numpy as np


def revolute(xyz, rpy, axis, qi):
    """Gives a casadi function for the quaternion. [xyz, w] form."""
    roll, pitch, yaw = rpy
    # Origin rotation from RPY ZYX convention
    cr = cs.cos(roll/2.0)
    sr = cs.sin(roll/2.0)
    cp = cs.cos(pitch/2.0)
    sp = cs.sin(pitch/2.0)
    cy = cs.cos(yaw/2.0)
    sy = cs.sin(yaw/2.0)

    # The quaternion associated with the origin rotation
    # Note: quat = [ xyz, w], where w is the scalar part
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    q_or = [x_or, y_or, z_or, w_or]
    # Joint rotation from axis angle
    cqi = cs.cos(qi/2.0)
    sqi = cs.sin(qi/2.0)
    x_j = axis[0]*sqi
    y_j = axis[1]*sqi
    z_j = axis[2]*sqi
    w_j = cqi
    q_j = [x_j, y_j, z_j, w_j]
    # Resulting quaternion
    return product(q_or, q_j)


def product(quat0, quat1):
    """Returns the quaternion product of q0 and q1."""
    quat = cs.SX.zeros(4)
    x0, y0, z0, w0 = quat0[0], quat0[1], quat0[2], quat0[3]
    x1, y1, z1, w1 = quat1[0], quat1[1], quat1[2], quat1[3]
    quat[0] = w0*x1 + x0*w1 + y0*z1 - z0*y1
    quat[1] = w0*y1 - x0*z1 + y0*w1 + z0*x1
    quat[2] = w0*z1 + x0*y1 - y0*x1 + z0*w1
    quat[3] = w0*w1 - x0*x1 - y0*y1 - z0*z1
    return quat


def numpy_rpy(roll, pitch, yaw):
    """Returns a quaternion ([x,y,z,w], w scalar) from roll pitch yaw ZYX
    convention."""
    cr = np.cos(roll/2.0)
    sr = np.sin(roll/2.0)
    cp = np.cos(pitch/2.0)
    sp = np.sin(pitch/2.0)
    cy = np.cos(yaw/2.0)
    sy = np.sin(yaw/2.0)
    x = sr * cp * cy - cr * sp * sy
    y = cr * sp * cy + sr * cp * sy
    z = cr * cp * sy - sr * sp * cy
    w = cr * cp * cy + sr * sp * sy
    # Remember to normalize:
    nq = np.sqrt(x*x + y*y + z*z + w*w)
    return np.array([x/nq,
                     y/nq,
                     z/nq,
                     w/nq])


def numpy_product(quat0, quat1):
    """Returns the quaternion product of q0 and q1."""
    quat = np.zeros(4)
    x0, y0, z0, w0 = quat0[0], quat0[1], quat0[2], quat0[3]
    x1, y1, z1, w1 = quat1[0], quat1[1], quat1[2], quat1[3]
    quat[0] = w0*x1 + x0*w1 + y0*z1 - z0*y1
    quat[1] = w0*y1 - x0*z1 + y0*w1 + z0*x1
    quat[2] = w0*z1 + x0*y1 - y0*x1 + z0*w1
    quat[3] = w0*w1 - x0*x1 - y0*y1 - z0*z1
    return quat


def numpy_ravani_roth_dist(q1, q2):
    """Quaternion distance designed by ravani and roth.
    See comparisons at: https://link.springer.com/content/pdf/10.1007%2Fs10851-009-0161-2.pdf"""
    return min(np.linalg.norm(q1 - q2), np.linalg.norm(q1 + q2))


def numpy_inner_product_dist(q1, q2):
    """Quaternion distance based on innerproduct.
    See comparisons at: https://link.springer.com/content/pdf/10.1007%2Fs10851-009-0161-2.pdf"""
    return 1.0 - abs(q1[0]*q2[0] + q1[1]*q2[1] + q1[2]*q2[2] + q1[3]*q2[3])



import casadi as ca


class FkCreator(object):
    def __init__(self, robotType, n):
        if robotType == 'panda':
            from forwardkinematics.urdfFks.pandaFk import PandaFk
            self._fk = PandaFk(n)
        elif robotType == 'boxer':
            from forwardkinematics.urdfFks.boxerFk import BoxerFk
            self._fk = BoxerFk(n)
        elif robotType == 'albert':
            from forwardkinematics.urdfFks.albertFk import AlbertFk
            self._fk = AlbertFk(n)
        elif robotType == 'tiago':
            from forwardkinematics.urdfFks.tiagoFk import TiagoFk
            self._fk = TiagoFk(n-4)
        elif robotType == 'planarArm':
            from forwardkinematics.planarFks.planarArmFk import PlanarArmFk
            self._fk = PlanarArmFk(n)
        elif robotType == 'mobilePanda':
            from forwardkinematics.urdfFks.mobilePandaFk import MobilePandaFk
            self._fk = MobilePandaFk(n)
        elif robotType == 'pointRobot' or robotType == 'pointMass':
            from forwardkinematics.planarFks.pointFk import PointFk
            self._fk = PointFk(n)
        elif robotType == 'groundRobot':
            from forwardkinematics.planarFks.groundRobotFk import GroundRobotFk
            self._fk = GroundRobotFk(n)

    def fk(self):
        return self._fk


if __name__ == "__main__":
    fk = FkCreator('boxer', 3).fk()
    q_ca = ca.SX.sym('q', fk.n())
    fk = fk.fk(q_ca, fk.n(), positionOnly=True)
    print(fk)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class BoxerFk(URDFForwardKinematics):

    def __init__(self, n):
        fileName = 'boxer.urdf'
        relevantLinks = ['world', 'ee_link', 'ee_link', 'ee_link']
        super().__init__(fileName, relevantLinks, 'world', 3)


if __name__ == "__main__":
    q_ca = ca.SX.sym("q_ca", 3)
    bfk = BoxerFk(3)
    fkCasadi = bfk.fk(q_ca, 0, positionOnly=True)
    print(fkCasadi)
    q_np = np.array([0.0, 1.0, 1 * np.pi/2.0])
    for i in range(bfk.n()):
        print(bfk.fk(q_np, i, positionOnly=True))



"""This module contains a class for turning a chain in a URDF to a
casadi function.
"""
import casadi as cs
import numpy as np
from platform import machine, system
from urdf_parser_py.urdf import URDF, Pose
import forwardkinematics.urdfFks.casadiConversion.geometry.transformation_matrix as T
import forwardkinematics.urdfFks.casadiConversion.geometry.plucker as plucker
import forwardkinematics.urdfFks.casadiConversion.geometry.quaternion as quaternion
import forwardkinematics.urdfFks.casadiConversion.geometry.dual_quaternion as dual_quaternion


class URDFparser(object):
    """Class that turns a chain from URDF to casadi functions."""
    actuated_types = ["prismatic", "revolute", "continuous"]
    func_opts = {"jit": True, "jit_options": {"flags": "-Ofast"}}
    # OS/CPU dependent specification of compiler
    if system() == "darwin" or machine() == "aarch64":
        func_opts["compiler"] = "shell"
    
    def __init__(self, func_opts=None):
        self.robot_desc = None
        if func_opts:
            self.func_opts = func_opts

    def from_file(self, filename):
        """Uses an URDF file to get robot description."""
        self.robot_desc = URDF.from_xml_file(filename)

    def from_server(self, key="robot_description"):
        """Uses a parameter server to get robot description."""
        self.robot_desc = URDF.from_parameter_server(key=key)

    def from_string(self, urdfstring):
        """Uses a URDF string to get robot description."""
        self.robot_desc = URDF.from_xml_string(urdfstring)

    def get_joint_info(self, root, tip):
        """Using an URDF to extract joint information, i.e list of
        joints, actuated names and upper and lower limits."""
        chain = self.robot_desc.get_chain(root, tip)
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        joint_list = []
        upper = []
        lower = []
        actuated_names = []

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]
                joint_list += [joint]
                if joint.type in self.actuated_types:
                    actuated_names += [joint.name]
                    if joint.type == "continuous":
                        upper += [cs.inf]
                        lower += [-cs.inf]
                    else:
                        upper += [joint.limit.upper]
                        lower += [joint.limit.lower]
                    if joint.axis is None:
                        joint.axis = [1., 0., 0.]
                    if joint.origin is None:
                        joint.origin = Pose(xyz=[0., 0., 0.],
                                            rpy=[0., 0., 0.])
                    elif joint.origin.xyz is None:
                        joint.origin.xyz = [0., 0., 0.]
                    elif joint.origin.rpy is None:
                        joint.origin.rpy = [0., 0., 0.]

        return joint_list, actuated_names, upper, lower

    def get_dynamics_limits(self, root, tip):
        """Using an URDF to extract joint max effort and velocity"""

        chain = self.robot_desc.get_chain(root, tip)
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        max_effort = []
        max_velocity = []

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]
                if joint.type in self.actuated_types:
                    if joint.limit is None:
                        max_effort += [cs.inf]
                        max_velocity += [cs.inf]
                    else:
                        max_effort += [joint.limit.effort]
                        max_velocity += [joint.limit.velocity]
        max_effort = [cs.inf if x is None else x for x in max_effort]
        max_velocity = [cs.inf if x is None else x for x in max_velocity]

        return max_effort, max_velocity

    def get_friction_matrices(self, root, tip):
        """Using an URDF to extract joint frictions and dampings"""

        chain = self.robot_desc.get_chain(root, tip)
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        friction = []
        damping = []

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]
                if joint.type in self.actuated_types:
                    if joint.dynamis is None:
                        friction += [0]
                        damping += [0]
                    else:
                        friction += [joint.dynamics.friction]
                        damping += [joint.dynamics.damping]
        friction = [0 if x is None else x for x in friction]
        damping = [0 if x is None else x for x in damping]
        Fv = np.diag(friction)
        Fd = np.diag(damping)
        return Fv, Fd


    def get_n_joints(self, root, tip):
        """Returns number of actuated joints."""

        chain = self.robot_desc.get_chain(root, tip)
        n_actuated = 0

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]
                if joint.type in self.actuated_types:
                    n_actuated += 1

        return n_actuated

    def _model_calculation(self, root, tip, q):
        """Calculates and returns model information needed in the
        dynamics algorithms caluculations, i.e transforms, joint space
        and inertia."""
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        chain = self.robot_desc.get_chain(root, tip)
        spatial_inertias = []
        i_X_0 = []
        i_X_p = []
        Sis = []
        prev_joint = None
        n_actuated = 0
        i = 0

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]

                if joint.type == "fixed":
                    if prev_joint == "fixed":
                        XT_prev = cs.mtimes(
                            plucker.XT(joint.origin.xyz, joint.origin.rpy),
                            XT_prev)
                    else:
                        XT_prev = plucker.XT(
                            joint.origin.xyz,
                            joint.origin.rpy)
                    inertia_transform = XT_prev
                    prev_inertia = spatial_inertia

                elif joint.type == "prismatic":
                    if n_actuated != 0:
                        spatial_inertias.append(spatial_inertia)
                    n_actuated += 1
                    XJT = plucker.XJT_prismatic(
                        joint.origin.xyz,
                        joint.origin.rpy,
                        joint.axis, q[i])
                    if prev_joint == "fixed":
                        XJT = cs.mtimes(XJT, XT_prev)
                    Si = cs.SX([0, 0, 0,
                                joint.axis[0],
                                joint.axis[1],
                                joint.axis[2]])
                    i_X_p.append(XJT)
                    Sis.append(Si)
                    i += 1

                elif joint.type in ["revolute", "continuous"]:
                    if n_actuated != 0:
                        spatial_inertias.append(spatial_inertia)
                    n_actuated += 1

                    XJT = plucker.XJT_revolute(
                        joint.origin.xyz,
                        joint.origin.rpy,
                        joint.axis,
                        q[i])
                    if prev_joint == "fixed":
                        XJT = cs.mtimes(XJT, XT_prev)
                    Si = cs.SX([
                                joint.axis[0],
                                joint.axis[1],
                                joint.axis[2],
                                0,
                                0,
                                0])
                    i_X_p.append(XJT)
                    Sis.append(Si)
                    i += 1

                prev_joint = joint.type

            if item in self.robot_desc.link_map:
                link = self.robot_desc.link_map[item]

                if link.inertial is None:
                    spatial_inertia = np.zeros((6, 6))
                else:
                    I = link.inertial.inertia
                    spatial_inertia = plucker.spatial_inertia_matrix_IO(
                        I.ixx,
                        I.ixy,
                        I.ixz,
                        I.iyy,
                        I.iyz,
                        I.izz,
                        link.inertial.mass,
                        link.inertial.origin.xyz)

                if prev_joint == "fixed":
                    spatial_inertia = prev_inertia + cs.mtimes(
                        inertia_transform.T,
                        cs.mtimes(spatial_inertia, inertia_transform))

                if link.name == tip:
                    spatial_inertias.append(spatial_inertia)

        return i_X_p, Sis, spatial_inertias

    def _apply_external_forces(self, external_f, f, i_X_p):
        """Internal function for applying external forces in dynamics
        algorithms calculations."""
        for i in range(0, len(f)):
            f[i] -= cs.mtimes(i_X_p[i].T, external_f[i])
        return f

    def get_inverse_dynamics_rnea(self, root, tip,
                                  gravity=None, f_ext=None):
        """Returns the inverse dynamics as a casadi function."""
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        q_dot = cs.SX.sym("q_dot", n_joints)
        q_ddot = cs.SX.sym("q_ddot", n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)

        v = []
        a = []
        f = []
        tau = cs.SX.zeros(n_joints)

        for i in range(0, n_joints):
            vJ = cs.mtimes(Si[i], q_dot[i])
            if i == 0:
                v.append(vJ)
                if gravity is not None:
                    ag = np.array([0.,
                                   0.,
                                   0.,
                                   gravity[0],
                                   gravity[1],
                                   gravity[2]])
                    a.append(
                        cs.mtimes(i_X_p[i], -ag) + cs.mtimes(Si[i], q_ddot[i]))
                else:
                    a.append(cs.mtimes(Si[i], q_ddot[i]))
            else:
                v.append(cs.mtimes(i_X_p[i], v[i-1]) + vJ)
                a.append(
                    cs.mtimes(i_X_p[i], a[i-1])
                    + cs.mtimes(Si[i], q_ddot[i])
                    + cs.mtimes(plucker.motion_cross_product(v[i]), vJ))

            f.append(
                cs.mtimes(Ic[i], a[i])
                + cs.mtimes(
                    plucker.force_cross_product(v[i]),
                    cs.mtimes(Ic[i], v[i])))

        if f_ext is not None:
            f = self._apply_external_forces(f_ext, f, i_X_p)

        for i in range(n_joints-1, -1, -1):
            tau[i] = cs.mtimes(Si[i].T, f[i])
            if i != 0:
                f[i-1] = f[i-1] + cs.mtimes(i_X_p[i].T, f[i])

        tau = cs.Function("C", [q, q_dot, q_ddot], [tau], self.func_opts)
        return tau

    def get_gravity_rnea(self, root, tip, gravity):
        """Returns the gravitational term as a casadi function."""

        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)

        v = []
        a = []
        ag = cs.SX([0., 0., 0., gravity[0], gravity[1], gravity[2]])
        f = []
        tau = cs.SX.zeros(n_joints)

        for i in range(0, n_joints):
            if i == 0:
                a.append(cs.mtimes(i_X_p[i], -ag))
            else:
                a.append(cs.mtimes(i_X_p[i], a[i-1]))
            f.append(cs.mtimes(Ic[i], a[i]))

        for i in range(n_joints-1, -1, -1):
            tau[i] = cs.mtimes(Si[i].T, f[i])
            if i != 0:
                f[i-1] = f[i-1] + cs.mtimes(i_X_p[i].T, f[i])

        tau = cs.Function("C", [q], [tau],
                          self.func_opts)
        return tau

    def _get_M(self, Ic, i_X_p, Si, n_joints, q):
        """Internal function for calculating the inertia matrix."""
        M = cs.SX.zeros(n_joints, n_joints)
        Ic_composite = [None]*len(Ic)

        for i in range(0, n_joints):
            Ic_composite[i] = Ic[i]

        for i in range(n_joints-1, -1, -1):
            if i != 0:
                Ic_composite[i-1] = (Ic[i-1]
                  + cs.mtimes(i_X_p[i].T,
                              cs.mtimes(Ic_composite[i], i_X_p[i])))

        for i in range(0, n_joints):
            fh = cs.mtimes(Ic_composite[i], Si[i])
            M[i, i] = cs.mtimes(Si[i].T, fh)
            j = i
            while j != 0:
                fh = cs.mtimes(i_X_p[j].T, fh)
                j -= 1
                M[i, j] = cs.mtimes(Si[j].T, fh)
                M[j, i] = M[i, j]

        return M

    def get_inertia_matrix_crba(self, root, tip):
        """Returns the inertia matrix as a casadi function."""
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)
        M = cs.SX.zeros(n_joints, n_joints)
        Ic_composite = [None]*len(Ic)

        for i in range(0, n_joints):
            Ic_composite[i] = Ic[i]

        for i in range(n_joints-1, -1, -1):
            if i != 0:
                Ic_composite[i-1] = Ic[i-1] + cs.mtimes(i_X_p[i].T, cs.mtimes(Ic_composite[i], i_X_p[i]))

        for i in range(0, n_joints):
            fh = cs.mtimes(Ic_composite[i], Si[i])
            M[i, i] = cs.mtimes(Si[i].T, fh)
            j = i
            while j != 0:
                fh = cs.mtimes(i_X_p[j].T, fh)
                j -= 1
                M[i, j] = cs.mtimes(Si[j].T, fh)
                M[j, i] = M[i, j]

        M = cs.Function("M", [q], [M], self.func_opts)
        return M

    def _get_C(self, i_X_p, Si, Ic, q, q_dot, n_joints,
               gravity=None, f_ext=None):
        """Internal function for calculating the joint space bias matrix."""

        v = []
        a = []
        f = []
        C = cs.SX.zeros(n_joints)

        for i in range(0, n_joints):
            vJ = cs.mtimes(Si[i], q_dot[i])
            if i == 0:
                v.append(vJ)
                if gravity is not None:
                    ag = np.array([0., 0., 0., gravity[0], gravity[1], gravity[2]])
                    a.append(cs.mtimes(i_X_p[i], -ag))
                else:
                    a.append(cs.SX([0., 0., 0., 0., 0., 0.]))
            else:
                v.append(cs.mtimes(i_X_p[i], v[i-1]) + vJ)
                a.append(cs.mtimes(i_X_p[i], a[i-1]) + cs.mtimes(plucker.motion_cross_product(v[i]),vJ))

            f.append(cs.mtimes(Ic[i], a[i]) + cs.mtimes(plucker.force_cross_product(v[i]), cs.mtimes(Ic[i], v[i])))

        if f_ext is not None:
            f = self._apply_external_forces(f_ext, f, i_X_0)

        for i in range(n_joints-1, -1, -1):
            C[i] = cs.mtimes(Si[i].T, f[i])
            if i != 0:
                f[i-1] = f[i-1] + cs.mtimes(i_X_p[i].T, f[i])

        return C

    def get_coriolis_rnea(self, root, tip, f_ext=None):
        """Returns the Coriolis matrix as a casadi function."""

        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        q_dot = cs.SX.sym("q_dot", n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)

        v = []
        a = []
        f = []
        tau = cs.SX.zeros(n_joints)

        for i in range(0, n_joints):
            vJ = cs.mtimes(Si[i], q_dot[i])

            if i == 0:
                v.append(vJ)
                a.append(cs.SX([0., 0., 0., 0., 0., 0.]))
            else:
                v.append(cs.mtimes(i_X_p[i], v[i-1]) + vJ)
                a.append(cs.mtimes(i_X_p[i], a[i-1]) + cs.mtimes(plucker.motion_cross_product(v[i]), vJ))

            f.append(cs.mtimes(Ic[i], a[i]) + cs.mtimes(plucker.force_cross_product(v[i]), cs.mtimes(Ic[i], v[i])))

        if f_ext is not None:
            f = self._apply_external_forces(f_ext, f, i_X_0)

        for i in range(n_joints-1, -1, -1):
            tau[i] = cs.mtimes(Si[i].T, f[i])
            if i != 0:
                f[i-1] = f[i-1] + cs.mtimes(i_X_p[i].T, f[i])

        C = cs.Function("C", [q, q_dot], [tau], self.func_opts)
        return C

    def get_forward_dynamics_crba(self, root, tip, gravity=None, f_ext=None):
        """Returns the forward dynamics as a casadi function by
        solving the Lagrangian eq. of motion.  OBS! Not appropriate
        for robots with a high number of dof -> use
        get_forward_dynamics_aba().
        """
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')
        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        q_dot = cs.SX.sym("q_dot", n_joints)
        tau = cs.SX.sym("tau", n_joints)
        q_ddot = cs.SX.zeros(n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)

        M = self._get_M(Ic, i_X_p, Si, n_joints, q)
        M_inv = cs.solve(M, cs.SX.eye(M.size1()))

        C = self._get_C(i_X_p, Si, Ic, q, q_dot, n_joints, gravity, f_ext)

        q_ddot = cs.mtimes(M_inv, (tau - C))
        q_ddot = cs.Function("q_ddot", [q, q_dot, tau],
                             [q_ddot], self.func_opts)

        return q_ddot

    def get_forward_dynamics_aba(self, root, tip, gravity=None, f_ext=None):
        """Returns the forward dynamics as a casadi function using the
        articulated body algorithm."""

        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        n_joints = self.get_n_joints(root, tip)
        q = cs.SX.sym("q", n_joints)
        q_dot = cs.SX.sym("q_dot", n_joints)
        tau = cs.SX.sym("tau", n_joints)
        q_ddot = cs.SX.zeros(n_joints)
        i_X_p, Si, Ic = self._model_calculation(root, tip, q)

        v = []
        c = []
        pA = []
        IA = []

        u = [None]*n_joints
        U = [None]*n_joints
        d = [None]*n_joints

        for i in range(0, n_joints):
            vJ = cs.mtimes(Si[i], q_dot[i])
            if i == 0:
                v.append(vJ)
                c.append([0, 0, 0, 0, 0, 0])
            else:
                v.append(cs.mtimes(i_X_p[i], v[i-1]) + vJ)
                c.append(cs.mtimes(plucker.motion_cross_product(v[i]), vJ))
            IA.append(Ic[i])
            pA.append(cs.mtimes(plucker.force_cross_product(v[i]),
                                cs.mtimes(Ic[i], v[i])))

        if f_ext is not None:
            pA = self._apply_external_forces(f_ext, pA)

        for i in range(n_joints-1, -1, -1):
            U[i] = cs.mtimes(IA[i], Si[i])
            d[i] = cs.mtimes(Si[i].T, U[i])
            u[i] = tau[i] - cs.mtimes(Si[i].T, pA[i])
            if i != 0:
                Ia = IA[i] - ((cs.mtimes(U[i], U[i].T)/d[i]))
                pa = pA[i] + cs.mtimes(Ia, c[i]) + (cs.mtimes(U[i], u[i])/d[i])
                IA[i-1] += cs.mtimes(i_X_p[i].T, cs.mtimes(Ia, i_X_p[i]))
                pA[i-1] += cs.mtimes(i_X_p[i].T, pa)

        a = []
        for i in range(0, n_joints):
            if i == 0:
                if gravity is not None:
                    ag = np.array([0.,
                                   0.,
                                   0.,
                                   gravity[0],
                                   gravity[1],
                                   gravity[2]])
                    a_temp = (cs.mtimes(i_X_p[i], -ag) + c[i])
                else:
                    a_temp = c[i]
            else:
                a_temp = (cs.mtimes(i_X_p[i], a[i-1]) + c[i])
            q_ddot[i] = (u[i] - cs.mtimes(U[i].T, a_temp))/d[i]
            a.append(a_temp + cs.mtimes(Si[i], q_ddot[i]))

        q_ddot = cs.Function("q_ddot", [q, q_dot, tau],
                             [q_ddot], self.func_opts)
        return q_ddot

    def get_forward_kinematics(self, root, tip, q):
        """Returns the forward kinematics as a casadi function."""
        chain = self.robot_desc.get_chain(root, tip)
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')
        joint_list, actuated_names, upper, lower = self.get_joint_info(
            root,
            tip)
        T_fk = cs.SX.eye(4)
        # nvar = len(actuated_names)
        # q = cs.SX.sym("q", nvar)
        quaternion_fk = cs.SX.zeros(4)
        quaternion_fk[3] = 1.0
        dual_quaternion_fk = cs.SX.zeros(8)
        dual_quaternion_fk[3] = 1.0
        i = 0
        for joint in joint_list:
            if joint.type == "fixed":
                xyz = joint.origin.xyz
                rpy = joint.origin.rpy
                joint_frame = T.numpy_rpy(xyz, *rpy)
                joint_quaternion = quaternion.numpy_rpy(*rpy)
                joint_dual_quat = dual_quaternion.numpy_prismatic(
                    xyz,
                    rpy,
                    [1., 0., 0.],
                    0.)
                T_fk = cs.mtimes(T_fk, joint_frame)
                quaternion_fk = quaternion.product(
                    quaternion_fk,
                    joint_quaternion)
                dual_quaternion_fk = dual_quaternion.product(
                    dual_quaternion_fk,
                    joint_dual_quat)

            elif joint.type == "prismatic":
                if joint.axis is None:
                    axis = cs.np.array([1., 0., 0.])
                else:
                    axis = cs.np.array(joint.axis)
                # axis = (1./cs.np.linalg.norm(axis))*axis
                joint_frame = T.prismatic(joint.origin.xyz,
                                          joint.origin.rpy,
                                          joint.axis, q[i])
                joint_quaternion = quaternion.numpy_rpy(*joint.origin.rpy)
                joint_dual_quat = dual_quaternion.prismatic(
                    joint.origin.xyz,
                    joint.origin.rpy,
                    axis, q[i])
                T_fk = cs.mtimes(T_fk, joint_frame)
                quaternion_fk = quaternion.product(quaternion_fk,
                                                   joint_quaternion)
                dual_quaternion_fk = dual_quaternion.product(
                    dual_quaternion_fk,
                    joint_dual_quat)
                i += 1

            elif joint.type in ["revolute", "continuous"]:
                if joint.axis is None:
                    axis = cs.np.array([1., 0., 0.])
                else:
                    axis = cs.np.array(joint.axis)
                axis = (1./cs.np.linalg.norm(axis))*axis
                joint_frame = T.revolute(
                    joint.origin.xyz,
                    joint.origin.rpy,
                    joint.axis, q[i])
                joint_quaternion = quaternion.revolute(
                    joint.origin.xyz,
                    joint.origin.rpy,
                    axis, q[i])
                joint_dual_quat = dual_quaternion.revolute(
                    joint.origin.xyz,
                    joint.origin.rpy,
                    axis, q[i])
                T_fk = cs.mtimes(T_fk, joint_frame)
                quaternion_fk = quaternion.product(
                    quaternion_fk,
                    joint_quaternion)
                dual_quaternion_fk = dual_quaternion.product(
                    dual_quaternion_fk,
                    joint_dual_quat)
                i += 1
        #T_fk = cs.Function("T_fk", [q], [T_fk], self.func_opts)
        quaternion_fk = cs.Function("quaternion_fk",
                                    [q], [quaternion_fk], self.func_opts)
        dual_quaternion_fk = cs.Function("dual_quaternion_fk",
                                         [q], [dual_quaternion_fk], self.func_opts)

        return {
            "joint_names": actuated_names,
            "upper": upper,
            "lower": lower,
            "joint_list": joint_list,
            "q": q,
            "quaternion_fk": quaternion_fk,
            "dual_quaternion_fk": dual_quaternion_fk,
            "T_fk": T_fk
        }



import casadi as cs
import forwardkinematics.urdfFks.casadiConversion.geometry.transformation_matrix as tm
import numpy as np


def numpy_skew_symmetric(v):
    """Returns a skew symmetric matrix from vector."""
    return np.array([[0, -v[2], v[1]],
                     [v[2], 0, -v[0]],
                     [-v[1], v[0], 0]])


def inertia_matrix(I):
    """Returns the 3x3 rotational inertia matrix given the inertia vector."""
    return np.array([I[0], I[1], I[2]],
                    [I[1], I[3], I[4]],
                    [I[2], I[4], I[5]])


def motion_cross_product(v):
    """Returns the motion cross product matrix of a spatial vector."""

    mcp = cs.SX.zeros(6, 6)

    mcp[0, 1] = -v[2]
    mcp[0, 2] = v[1]
    mcp[1, 0] = v[2]
    mcp[1, 2] = -v[0]
    mcp[2, 0] = -v[1]
    mcp[2, 1] = v[0]

    mcp[3, 4] = -v[2]
    mcp[3, 5] = v[1]
    mcp[4, 3] = v[2]
    mcp[4, 5] = -v[0]
    mcp[5, 3] = -v[1]
    mcp[5, 4] = v[0]

    mcp[3, 1] = -v[5]
    mcp[3, 2] = v[4]
    mcp[4, 0] = v[5]
    mcp[4, 2] = -v[3]
    mcp[5, 0] = -v[4]
    mcp[5, 1] = v[3]

    return mcp


def force_cross_product(v):
    """Returns the force cross product matrix of a spatial vector."""
    return -motion_cross_product(v).T


def spatial_inertia_matrix_Ic(ixx, ixy, ixz, iyy, iyz, izz, mass):
    """Returns the 6x6 spatial inertia matrix expressed at the center of
    mass."""
    Ic = np.zeros([6, 6])
    Ic[:3, :3] = np.array([[ixx, ixy, ixz],
                           [ixy, iyy, iyz],
                           [ixz, iyz, izz]])

    Ic[3, 3] = mass
    Ic[4, 4] = mass
    Ic[5, 5] = mass

    return Ic


def spatial_inertia_matrix_IO(ixx, ixy, ixz, iyy, iyz, izz, mass, c):
    """Returns the 6x6 spatial inertia matrix expressed at the origin."""
    IO = np.zeros([6, 6])
    cx = numpy_skew_symmetric(c)
    inertia_matrix = np.array([[ixx, ixy, ixz],
                               [ixy, iyy, iyz],
                               [ixz, iyz, izz]])

    IO[:3, :3] = inertia_matrix + mass*(np.dot(cx, np.transpose(cx)))
    IO[:3, 3:] = mass*cx
    IO[3:, :3] = mass*np.transpose(cx)

    IO[3, 3] = mass
    IO[4, 4] = mass
    IO[5, 5] = mass

    return IO


def numpy_rotation_rpy(roll, pitch, yaw):
    """Returns a rotation matrix from roll pitch yaw. ZYX convention."""
    cr = np.cos(roll)
    sr = np.sin(roll)
    cp = np.cos(pitch)
    sp = np.sin(pitch)
    cy = np.cos(yaw)
    sy = np.sin(yaw)
    return np.array([[cy*cp,  cy*sp*sr - sy*cr,  cy*sp*cr + sy*sr],
                     [sy*cp,  sy*sp*sr + cy*cr,  sy*sp*cr - cy*sr],
                     [  -sp,             cp*sr,             cp*cr]])


def spatial_force_transform(R, r):
    """Returns the spatial force transform from a 3x3 rotation matrix
    and a 3x1 displacement vector."""
    X = cs.SX.zeros(6, 6)
    X[:3, :3] = R.T
    X[3:, 3:] = R.T
    X[:3, 3:] = cs.mtimes(cs.skew(r), R.T)
    return X


def spatial_transform(R, r):
    """Returns the spatial motion transform from a 3x3 rotation matrix
    and a 3x1 displacement vector."""
    X = cs.SX.zeros(6, 6)
    X[:3, :3] = R
    X[3:, 3:] = R
    X[3:, :3] = -cs.mtimes(R, cs.skew(r))
    return X


def spatial_transform_BA(R, r):
    """Returns the inverse spatial motion transform from a 3x3 rotation
    matrix and a 3x1 displacement vector."""
    X = cs.SX.zeros(6, 6)
    X[:3, :3] = R.T
    X[3:, 3:] = R.T
    X[3:, :3] = cs.mtimes(cs.skew(r), R.T)
    return X


def XJT_revolute(xyz, rpy, axis, qi):
    """Returns the spatial transform from child link to parent link with
    a revolute connecting joint."""
    T = tm.revolute(xyz, rpy, axis, qi)
    rotation_matrix = T[:3, :3]
    displacement = T[:3, 3]
    return spatial_transform(rotation_matrix.T, displacement)


def XJT_revolute_BA(xyz, rpy, axis, qi):
    """Returns the spatial transform from parent link to child link with
    a revolute connecting joint."""
    T = tm.revolute(xyz, rpy, axis, qi)
    rotation_matrix = T[:3, :3]
    displacement = T[:3, 3]
    return spatial_transform_BA(rotation_matrix, displacement)


def XJT_prismatic(xyz, rpy, axis, qi):
    """Returns the spatial transform from child link to parent link with
    a prismatic connecting joint."""
    T = tm.prismatic(xyz, rpy, axis, qi)
    rotation_matrix = T[:3, :3]
    displacement = T[:3, 3]
    return spatial_transform(rotation_matrix.T, displacement)


def XJT_prismatic_BA(xyz, rpy, axis, qi):
    """Returns the spatial transform from parent link to child link with
    a prismatic connecting joint."""
    T = tm.prismatic(xyz, rpy, axis, qi)
    rotation_matrix = T[:3, :3]
    displacement = T[:3, 3]
    return spatial_transform_BA(rotation_matrix, displacement)


def XT(xyz, rpy):
    """Returns a general spatial transformation matrix matrix"""
    rotation_matrix = numpy_rotation_rpy(rpy[0], rpy[1], rpy[2])
    return spatial_transform(rotation_matrix.T, xyz)



import numpy as np
import casadi as ca
from forwardkinematics.fksCommon.fk import ForwardKinematics


class PlanarArmFk(ForwardKinematics):

    def __init__(self, n):
        super().__init__()
        self._n = n

    def fk(self, q, i, positionOnly=False, endlink=0.0):
        assert i <= self._n
        if isinstance(q, ca.SX):
            assert(q.shape[0] == self._n)
            return self.casadi(q, i, positionOnly=positionOnly, endlink=endlink)
        elif isinstance(q, np.ndarray):
            assert(q.size == self._n)
            return self.numpy(q, i, positionOnly=positionOnly, endlink=endlink)

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.SX(np.array([0.0, 0.0, q[0]]))
        for i in range(1, i + 1):
            fk[0] += ca.cos(fk[2]) * 1.0
            fk[1] += ca.sin(fk[2]) * 1.0
            if i < q.size(1):
                fk[2] += q[i]
        fk[0] += ca.cos(fk[2]) * endlink
        fk[1] += ca.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.array([0.0, 0.0, q[0]])
        for i in range(1, i+1):
            fk[0] += np.cos(fk[2]) * 1.0
            fk[1] += np.sin(fk[2]) * 1.0
            if i < len(q):
                fk[2] += q[i]
        fk[0] += np.cos(fk[2]) * endlink
        fk[1] += np.sin(fk[2]) * endlink
        if positionOnly:
            return fk[0:2]
        else:
            return fk


if __name__ == "__main__":
    q_ca = ca.SX.sym("q", 6)
    fkPlanar = PlanarArmFk(6)
    q_np = np.array([3.0, 0.0, 0.2, 0.0, 0.0, 0.2])
    fkCasadi = fkPlanar.fk(q_ca, 6, positionOnly=True)
    fkNumpy = fkPlanar.fk(q_np, 6, positionOnly=True)
    print(fkNumpy)
    print(fkCasadi)




import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class AlbertFk(URDFForwardKinematics):

    def __init__(self, n):
        fileName = 'albert.urdf'
        relevantLinks = ['world', 'base_tip_link', 'top_mount'] + ['panda_link' + str(i) for i in [0, 3, 4, 5, 6, 7, 8, 9]]
        super().__init__(fileName, relevantLinks, 'world', len(relevantLinks) - 1)


if __name__ == "__main__":
    q_ca = ca.SX.sym("q_ca", 10)
    pfk = AlbertFk(10)
    fkCasadi = pfk.fk(q_ca, 10, positionOnly=True)
    q_np = np.array([-0.01026582, -0.28908, 0.00922652, -1.3468, -0.026000, 2.074019, 0.06062268, 0.0, 0.0, 0.0])
    q_np *= 0
    fkNumpy = pfk.fk(q_np, 1, positionOnly=True)
    print(fkNumpy)



"""Functions for getting casadi expressions for dual quaternions from joint
type."""
import casadi as cs
import numpy as np


def product(Q, P):
    """Returns the dual quaternion product of two 8 element vectors
    representing a dual quaternions. First four elements are the real
    part, last four elements are the dual part.
    """
    res = cs.SX.zeros(8)
    # Real and dual components
    xr0, yr0, zr0, wr0 = Q[0], Q[1], Q[2], Q[3]
    xd0, yd0, zd0, wd0 = Q[4], Q[5], Q[6], Q[7]
    xr1, yr1, zr1, wr1 = P[0], P[1], P[2], P[3]
    xd1, yd1, zd1, wd1 = P[4], P[5], P[6], P[7]
    # Real part
    xr = wr0*xr1 + xr0*wr1 + yr0*zr1 - zr0*yr1
    yr = wr0*yr1 - xr0*zr1 + yr0*wr1 + zr0*xr1
    zr = wr0*zr1 + xr0*yr1 - yr0*xr1 + zr0*wr1
    wr = wr0*wr1 - xr0*xr1 - yr0*yr1 - zr0*zr1
    # Dual part
    xd = xr0*wd1 + wr0*xd1 + yr0*zd1 - zr0*yd1
    xd += xd0*wr1 + wd0*xr1 + yd0*zr1 - zd0*yr1
    yd = wr0*yd1 - xr0*zd1 + yr0*wd1 + zr0*xd1
    yd += wd0*yr1 - xd0*zr1 + yd0*wr1 + zd0*xr1
    zd = wr0*zd1 + xr0*yd1 - yr0*xd1 + zr0*wd1
    zd += wd0*zr1 + xd0*yr1 - yd0*xr1 + zd0*wr1
    wd = wr1*wd0 - xr1*xd0 - yr1*yd0 - zr1*zd0
    wd += wd1*wr0 - xd1*xr0 - yd1*yr0 - zd1*zr0

    res[0] = xr
    res[1] = yr
    res[2] = zr
    res[3] = wr
    res[4] = xd
    res[5] = yd
    res[6] = zd
    res[7] = wd
    return res


def conj(Q):
    """Returns the conjugate of a dual quaternion.
    """
    res = cs.SX.zeros(8)
    res[0] = -Q[0]
    res[1] = -Q[1]
    res[2] = -Q[2]
    res[3] = Q[3]
    res[4] = -Q[4]
    res[5] = -Q[5]
    res[6] = -Q[6]
    res[7] = Q[7]
    return res


def norm2(Q):
    """Returns the dual norm of a dual quaternion.
    Based on:
    https://github.com/bobbens/libdq/blob/master/dq.c
    """
    real = cs.SX.zeros(1)
    dual = cs.SX.zeros(1)
    real = Q[0]*Q[0] + Q[1]*Q[1] + Q[2]*Q[2] + Q[3]*Q[3]
    dual = 2.*(Q[3]*Q[7] + Q[0]*Q[4] + Q[1]*Q[5] + Q[2]*Q[6])
    return real, dual


def inv(Q):
    """Returns the inverse of a dual quaternion.
    Based on:
    https://github.com/bobbens/libdq/blob/master/dq.c
    """
    res = cs.SX.zeros(8)
    real, dual = norm2(Q)
    res[0] = -Q[0] * real
    res[1] = -Q[1] * real
    res[2] = -Q[2] * real
    res[3] = Q[3] * real
    res[4] = Q[4] * (dual-real)
    res[5] = Q[5] * (dual-real)
    res[6] = Q[6] * (dual-real)
    res[7] = Q[7] * (real-dual)
    return res


def to_transformation_matrix(Q):
    """Transforms a dual quaternion to a 4x4 transformation matrix.
    """
    res = cs.SX.zeros(4, 4)
    # Rotation part:
    xr, yr, zr, wr = Q[0], Q[1], Q[2], Q[3]
    xd, yd, zd, wd = Q[4], Q[5], Q[6], Q[7]
    res[0, 0] = wr*wr + xr*xr - yr*yr - zr*zr
    res[1, 1] = wr*wr - xr*xr + yr*yr - zr*zr
    res[2, 2] = wr*wr - xr*xr - yr*yr + zr*zr
    res[0, 1] = 2.*(xr*yr - wr*zr)
    res[1, 0] = 2.*(xr*yr + wr*zr)
    res[0, 2] = 2.*(xr*zr + wr*yr)
    res[2, 0] = 2.*(xr*zr - wr*yr)
    res[1, 2] = 2.*(yr*zr - wr*xr)
    res[2, 1] = 2.*(yr*zr + wr*xr)

    # Displacement part:
    res[0, 3] = 2.*(-wd*xr + xd*wr - yd*zr + zd*yr)
    res[1, 3] = 2.*(-wd*yr + xd*zr + yd*wr - zd*xr)
    res[2, 3] = 2.*(-wd*zr - xd*yr + yd*xr + zd*wr)
    res[3, 3] = 1.0
    return res


def to_rotation_matrix(Q):
    """Transforms a dual quaternion to 3x3 rotation matrix
    """
    res = cs.SX.zeros(3, 3)
    xr, yr, zr, wr = Q[0], Q[1], Q[2], Q[3]
    res[0, 0] = wr*wr + xr*xr - yr*yr - zr*zr
    res[1, 1] = wr*wr - xr*xr + yr*yr - zr*zr
    res[2, 2] = wr*wr - xr*xr - yr*yr + zr*zr
    res[0, 1] = 2.*(xr*yr - wr*zr)
    res[1, 0] = 2.*(xr*yr + wr*zr)
    res[0, 2] = 2.*(xr*zr + wr*yr)
    res[2, 0] = 2.*(xr*zr - wr*yr)
    res[1, 2] = 2.*(yr*zr - wr*xr)
    res[2, 1] = 2.*(yr*zr + wr*xr)
    return res


def to_position(Q):
    """Transforms a dual quaternion to a position.
    """
    res = cs.SX.zeros(3)
    xr, yr, zr, wr = Q[0], Q[1], Q[2], Q[3]
    xd, yd, zd, wd = Q[4], Q[5], Q[6], Q[7]
    res[0] = 2.*(-wd*xr + xd*wr - yd*zr + zd*yr)
    res[1] = 2.*(-wd*yr + xd*zr + yd*wr - zd*xr)
    res[2] = 2.*(-wd*zr - xd*yr + yd*xr + zd*wr)
    return res


def rpy(rpy):
    """Returns the dual quaternion for a pure roll-pitch-yaw rotation.
    """
    roll, pitch, yaw = rpy[0], rpy[1], rpy[2]
    # Origin rotation from RPY ZYX convention
    cr = cs.cos(roll/2.0)
    sr = cs.sin(roll/2.0)
    cp = cs.cos(pitch/2.0)
    sp = cs.sin(pitch/2.0)
    cy = cs.cos(yaw/2.0)
    sy = cs.sin(yaw/2.0)

    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    res = cs.SX.zeros(8)
    # Note, our dual quaternions use a different representation
    # dual_quat = [xyz, w, xyz', w']
    # where w + xyz represents the "real" quaternion
    # and w'+xyz' represents the "dual" quaternion
    res[0] = x_or
    res[1] = y_or
    res[2] = z_or
    res[3] = w_or
    return res


def translation(xyz):
    """Returns the dual quaternion for a pure translation.
    """
    res = cs.SX.zeros(8)
    res[3] = 1.0
    res[4] = xyz[0]/2.0
    res[5] = xyz[1]/2.0
    res[6] = xyz[2]/2.0
    return res


def axis_translation(axis, qi):
    """Returns the dual quaternion for a translation along an axis.
    """
    res = cs.SX.zeros(8)
    res[3] = 1.0
    res[4] = qi*axis[0]/2.0
    res[5] = qi*axis[1]/2.0
    res[6] = qi*axis[2]/2.0
    return res


def axis_rotation(axis, qi):
    """Returns the dual quaternion for a rotation along an axis.
    AXIS MUST BE NORMALIZED!
    """
    res = cs.SX.zeros(8)
    cqi = cs.cos(qi/2.0)
    sqi = cs.sin(qi/2.0)
    res[0] = axis[0]*sqi
    res[1] = axis[1]*sqi
    res[2] = axis[2]*sqi
    res[3] = cqi
    return res


def prismatic(xyz, rpy, axis, qi):
    """Returns the dual quaternion for a prismatic joint.
    """
    # Joint origin rotation from RPY ZYX convention
    roll, pitch, yaw = rpy[0], rpy[1], rpy[2]
    # Origin rotation from RPY ZYX convention
    cr = cs.cos(roll/2.0)
    sr = cs.sin(roll/2.0)
    cp = cs.cos(pitch/2.0)
    sp = cs.sin(pitch/2.0)
    cy = cs.cos(yaw/2.0)
    sy = cs.sin(yaw/2.0)
    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    # Joint origin translation as a dual quaternion
    x_ot = 0.5*xyz[0]*w_or + 0.5*xyz[1]*z_or - 0.5*xyz[2]*y_or
    y_ot = - 0.5*xyz[0]*z_or + 0.5*xyz[1]*w_or + 0.5*xyz[2]*x_or
    z_ot = 0.5*xyz[0]*y_or - 0.5*xyz[1]*x_or + 0.5*xyz[2]*w_or
    w_ot = - 0.5*xyz[0]*x_or - 0.5*xyz[1]*y_or - 0.5*xyz[2]*z_or
    Q_o = [x_or, y_or, z_or, w_or, x_ot, y_ot, z_ot, w_ot]
    # Joint displacement orientation is just identity
    x_jr = 0.0
    y_jr = 0.0
    z_jr = 0.0
    w_jr = 1.0
    # Joint displacement translation along axis
    x_jt = qi*axis[0]/2.0
    y_jt = qi*axis[1]/2.0
    z_jt = qi*axis[2]/2.0
    w_jt = 0.0
    Q_j = [x_jr, y_jr, z_jr, w_jr, x_jt, y_jt, z_jt, w_jt]
    # Get resulting dual quaternion
    return product(Q_o, Q_j)


def revolute(xyz, rpy, axis, qi):
    """Returns the dual quaternion for a revolute joint.
    AXIS MUST BE NORMALIZED!
    """
    # Joint origin rotation from RPY ZYX convention
    roll, pitch, yaw = rpy[0], rpy[1], rpy[2]
    # Origin rotation from RPY ZYX convention
    cr = cs.cos(roll/2.0)
    sr = cs.sin(roll/2.0)
    cp = cs.cos(pitch/2.0)
    sp = cs.sin(pitch/2.0)
    cy = cs.cos(yaw/2.0)
    sy = cs.sin(yaw/2.0)
    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    # Joint origin translation as a dual quaternion
    x_ot = 0.5*xyz[0]*w_or + 0.5*xyz[1]*z_or - 0.5*xyz[2]*y_or
    y_ot = - 0.5*xyz[0]*z_or + 0.5*xyz[1]*w_or + 0.5*xyz[2]*x_or
    z_ot = 0.5*xyz[0]*y_or - 0.5*xyz[1]*x_or + 0.5*xyz[2]*w_or
    w_ot = - 0.5*xyz[0]*x_or - 0.5*xyz[1]*y_or - 0.5*xyz[2]*z_or
    Q_o = [x_or, y_or, z_or, w_or, x_ot, y_ot, z_ot, w_ot]
    # Joint displacement rotation is from axis angle
    # nax = cs.sqrt(axis[0]*axis[0] + axis[1]*axis[1] + axis[2]*axis[2])
    cqi = cs.cos(qi/2.0)
    sqi = cs.sin(qi/2.0)
    x_jr = axis[0]*sqi
    y_jr = axis[1]*sqi
    z_jr = axis[2]*sqi
    w_jr = cqi
    # Joint displacement translation is nothing
    x_jt = 0.0
    y_jt = 0.0
    z_jt = 0.0
    w_jt = 0.0
    Q_j = [x_jr, y_jr, z_jr, w_jr, x_jt, y_jt, z_jt, w_jt]
    return product(Q_o, Q_j)


def numpy_product(Q, P):
    """Returns the dual quaternion product of two 8 element vectors
    representing a dual quaternions. First four elements are the real
    part, last four elements are the dual part.
    """
    res = np.zeros(8)
    # Real and dual components
    xr0, yr0, zr0, wr0 = Q[0], Q[1], Q[2], Q[3]
    xd0, yd0, zd0, wd0 = Q[4], Q[5], Q[6], Q[7]
    xr1, yr1, zr1, wr1 = P[0], P[1], P[2], P[3]
    xd1, yd1, zd1, wd1 = P[4], P[5], P[6], P[7]
    # Real part
    xr = wr0*xr1 + xr0*wr1 + yr0*zr1 - zr0*yr1
    yr = wr0*yr1 - xr0*zr1 + yr0*wr1 + zr0*xr1
    zr = wr0*zr1 + xr0*yr1 - yr0*xr1 + zr0*wr1
    wr = wr0*wr1 - xr0*xr1 - yr0*yr1 - zr0*zr1
    # Dual part
    xd = xr0*wd1 + wr0*xd1 + yr0*zd1 - zr0*yd1
    xd += xd0*wr1 + wd0*xr1 + yd0*zr1 - zd0*yr1
    yd = wr0*yd1 - xr0*zd1 + yr0*wd1 + zr0*xd1
    yd += wd0*yr1 - xd0*zr1 + yd0*wr1 + zd0*xr1
    zd = wr0*zd1 + xr0*yd1 - yr0*xd1 + zr0*wd1
    zd += wd0*zr1 + xd0*yr1 - yd0*xr1 + zd0*wr1
    wd = wr1*wd0 - xr1*xd0 - yr1*yd0 - zr1*zd0
    wd += wd1*wr0 - xd1*xr0 - yd1*yr0 - zd1*zr0

    res[0] = xr
    res[1] = yr
    res[2] = zr
    res[3] = wr
    res[4] = xd
    res[5] = yd
    res[6] = zd
    res[7] = wd
    return res


def numpy_conj(Q):
    """Returns the conjugate of a dual quaternion.
    """
    res = np.zeros(8)
    res[0] = -Q[0]
    res[1] = -Q[1]
    res[2] = -Q[2]
    res[3] = Q[3]
    res[4] = -Q[4]
    res[5] = -Q[5]
    res[6] = -Q[6]
    res[7] = Q[7]
    return res


def numpy_norm2(Q):
    """Returns the dual norm of a dual quaternion.
    Based on:
    https://github.com/bobbens/libdq/blob/master/dq.c
    """
    real = Q[0]*Q[0] + Q[1]*Q[1] + Q[2]*Q[2] + Q[3]*Q[3]
    dual = 2.*(Q[3]*Q[7] + Q[0]*Q[4] + Q[1]*Q[5] + Q[2]*Q[6])
    return real, dual


def numpy_inv(Q):
    """Returns the inverse of a dual quaternion.
    Based on:
    https://github.com/bobbens/libdq/blob/master/dq.c
    """
    res = np.zeros(8)
    real, dual = norm2(Q)
    res[0] = -Q[0] * real
    res[1] = -Q[1] * real
    res[2] = -Q[2] * real
    res[3] = Q[3] * real
    res[4] = Q[4] * (dual-real)
    res[5] = Q[5] * (dual-real)
    res[6] = Q[6] * (dual-real)
    res[7] = Q[7] * (real-dual)
    return res


def to_numpy_transformation_matrix(Q):
    """Transforms a dual quaternion to a 4x4 transformation matrix.
    """
    res = np.zeros((4, 4))
    # Rotation part:
    xr, yr, zr, wr = Q[0], Q[1], Q[2], Q[3]
    xd, yd, zd, wd = Q[4], Q[5], Q[6], Q[7]
    res[0, 0] = wr*wr + xr*xr - yr*yr - zr*zr
    res[1, 1] = wr*wr - xr*xr + yr*yr - zr*zr
    res[2, 2] = wr*wr - xr*xr - yr*yr + zr*zr
    res[0, 1] = 2.*(xr*yr - wr*zr)
    res[1, 0] = 2.*(xr*yr + wr*zr)
    res[0, 2] = 2.*(xr*zr + wr*yr)
    res[2, 0] = 2.*(xr*zr - wr*yr)
    res[1, 2] = 2.*(yr*zr - wr*xr)
    res[2, 1] = 2.*(yr*zr + wr*xr)

    # Displacement part:
    res[0, 3] = 2.*(-wd*xr + xd*wr - yd*zr + zd*yr)
    res[1, 3] = 2.*(-wd*yr + xd*zr + yd*wr - zd*xr)
    res[2, 3] = 2.*(-wd*zr - xd*yr + yd*xr + zd*wr)
    res[3, 3] = 1.0
    return res


def numpy_rpy(rpy):
    """Returns the dual quaternion for a pure roll-pitch-yaw rotation.
    """
    roll, pitch, yaw = rpy
    # Origin rotation from RPY ZYX convention
    cr = np.cos(roll/2.0)
    sr = np.sin(roll/2.0)
    cp = np.cos(pitch/2.0)
    sp = np.sin(pitch/2.0)
    cy = np.cos(yaw/2.0)
    sy = np.sin(yaw/2.0)

    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    res = np.zeros(8)
    # Note, our dual quaternions use a different representation
    # dual_quat = [xyz, w, xyz', w']
    # where w + xyz represents the "real" quaternion
    # and w'+xyz' represents the "dual" quaternion
    res[0] = x_or
    res[1] = y_or
    res[2] = z_or
    res[3] = w_or
    return res


def numpy_translation(xyz):
    """Returns the dual quaternion for a pure translation.
    """
    res = np.zeros(8)
    res[3] = 1.0
    res[4] = xyz[0]/2.0
    res[5] = xyz[1]/2.0
    res[6] = xyz[2]/2.0
    return res


def numpy_axis_translation(axis, qi):
    """Returns the dual quaternion for a translation along an axis.
    """
    res = np.zeros(8)
    res[3] = 1.0
    res[4] = qi*axis[0]/2.0
    res[5] = qi*axis[1]/2.0
    res[6] = qi*axis[2]/2.0
    return res


def numpy_axis_rotation(axis, qi):
    """Returns the dual quaternion for a rotation along an axis.
    AXIS MUST BE NORMALIZED!
    """
    res = np.zeros(8)
    cqi = np.cos(qi/2.0)
    sqi = np.sin(qi/2.0)
    res[0] = axis[0]*sqi
    res[1] = axis[1]*sqi
    res[2] = axis[2]*sqi
    res[3] = cqi
    return res


def numpy_prismatic(xyz, rpy, axis, qi):
    """Returns the dual quaternion for a prismatic joint.
    """
    # Joint origin rotation from RPY ZYX convention
    roll, pitch, yaw = rpy
    # Origin rotation from RPY ZYX convention
    cr = np.cos(roll/2.0)
    sr = np.sin(roll/2.0)
    cp = np.cos(pitch/2.0)
    sp = np.sin(pitch/2.0)
    cy = np.cos(yaw/2.0)
    sy = np.sin(yaw/2.0)
    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    # Joint origin translation as a dual quaternion
    x_ot = 0.5*xyz[0]*w_or + 0.5*xyz[1]*z_or - 0.5*xyz[2]*y_or
    y_ot = - 0.5*xyz[0]*z_or + 0.5*xyz[1]*w_or + 0.5*xyz[2]*x_or
    z_ot = 0.5*xyz[0]*y_or - 0.5*xyz[1]*x_or + 0.5*xyz[2]*w_or
    w_ot = - 0.5*xyz[0]*x_or - 0.5*xyz[1]*y_or - 0.5*xyz[2]*z_or
    Q_o = [x_or, y_or, z_or, w_or, x_ot, y_ot, z_ot, w_ot]
    # Joint displacement orientation is just identity
    x_jr = 0.0
    y_jr = 0.0
    z_jr = 0.0
    w_jr = 1.0
    # Joint displacement translation along axis
    x_jt = qi*axis[0]/2.0
    y_jt = qi*axis[1]/2.0
    z_jt = qi*axis[2]/2.0
    w_jt = 0.0
    Q_j = [x_jr, y_jr, z_jr, w_jr, x_jt, y_jt, z_jt, w_jt]
    # Get resulting dual quaternion
    return product(Q_o, Q_j)


def numpy_revolute(xyz, rpy, axis, qi):
    """Returns the dual quaternion for a revolute joint.
    AXIS MUST BE NORMALIZED!
    """
    # Joint origin rotation from RPY ZYX convention
    roll, pitch, yaw = rpy
    # Origin rotation from RPY ZYX convention
    cr = np.cos(roll/2.0)
    sr = np.sin(roll/2.0)
    cp = np.cos(pitch/2.0)
    sp = np.sin(pitch/2.0)
    cy = np.cos(yaw/2.0)
    sy = np.sin(yaw/2.0)
    # The quaternion associated with the origin rotation
    # Note: quat = w + ix + jy + kz
    x_or = cy*sr*cp - sy*cr*sp
    y_or = cy*cr*sp + sy*sr*cp
    z_or = sy*cr*cp - cy*sr*sp
    w_or = cy*cr*cp + sy*sr*sp
    # Joint origin translation as a dual quaternion
    x_ot = 0.5*xyz[0]*w_or + 0.5*xyz[1]*z_or - 0.5*xyz[2]*y_or
    y_ot = - 0.5*xyz[0]*z_or + 0.5*xyz[1]*w_or + 0.5*xyz[2]*x_or
    z_ot = 0.5*xyz[0]*y_or - 0.5*xyz[1]*x_or + 0.5*xyz[2]*w_or
    w_ot = - 0.5*xyz[0]*x_or - 0.5*xyz[1]*y_or - 0.5*xyz[2]*z_or
    Q_o = [x_or, y_or, z_or, w_or, x_ot, y_ot, z_ot, w_ot]
    # Joint displacement rotation is from axis angle
    cqi = np.cos(qi/2.0)
    sqi = np.sin(qi/2.0)
    x_jr = axis[0]*sqi
    y_jr = axis[1]*sqi
    z_jr = axis[2]*sqi
    w_jr = cqi
    # Joint displacement translation is nothing
    x_jt = 0.0
    y_jt = 0.0
    z_jt = 0.0
    w_jt = 0.0
    Q_j = [x_jr, y_jr, z_jr, w_jr, x_jt, y_jt, z_jt, w_jt]
    return product(Q_o, Q_j)



import numpy as np
import casadi as ca
from forwardkinematics.urdfFks.urdfFk import URDFForwardKinematics


class TiagoFk(URDFForwardKinematics):

    def __init__(self, n):
        fileName = 'tiago.urdf'
        linkIndices = [1, 2, 3, 4, 5, 7]
        relevantLinks = ['base_link', 'torso_lift_link']
        relevantLinks += ['arm_left_' + str(i) + '_link' for i in linkIndices] + ['arm_left_tool_link']
        relevantLinks += ['arm_right_' + str(i) + '_link' for i in linkIndices] + ['arm_left_tool_link']
        relevantLinks += ['head_1_link', 'head_2_link']
        super().__init__(fileName, relevantLinks, 'base_link', n)


if __name__ == "__main__":
    n = 17
    q_ca = ca.SX.sym("q_ca", n)
    pfk = TiagoFk(n)
    fkCasadi = pfk.fk(q_ca, 15, positionOnly=False)
    q_np = np.zeros(n)
    q_np[13] = 0.1
    for i in range(n+1):
        fkNumpy = pfk.fk(q_np, i, positionOnly=True)
        print(fkNumpy)



"""Functions for getting casadi expressions for transformation matrices from
joint type."""
import casadi as cs
import numpy as np


def prismatic(xyz, rpy, axis, qi):
    T = cs.SX.zeros(4, 4)

    # Origin rotation from RPY ZYX convention
    cr = cs.cos(rpy[0])
    sr = cs.sin(rpy[0])
    cp = cs.cos(rpy[1])
    sp = cs.sin(rpy[1])
    cy = cs.cos(rpy[2])
    sy = cs.sin(rpy[2])
    r00 = cy*cp
    r01 = cy*sp*sr - sy*cr
    r02 = cy*sp*cr + sy*sr
    r10 = sy*cp
    r11 = sy*sp*sr + cy*cr
    r12 = sy*sp*cr - cy*sr
    r20 = -sp
    r21 = cp*sr
    r22 = cp*cr
    p0 = r00*axis[0]*qi + r01*axis[1]*qi + r02*axis[2]*qi
    p1 = r10*axis[0]*qi + r11*axis[1]*qi + r12*axis[2]*qi
    p2 = r20*axis[0]*qi + r21*axis[1]*qi + r22*axis[2]*qi

    # Homogeneous transformation matrix
    T[0, 0] = r00
    T[0, 1] = r01
    T[0, 2] = r02
    T[1, 0] = r10
    T[1, 1] = r11
    T[1, 2] = r12
    T[2, 0] = r20
    T[2, 1] = r21
    T[2, 2] = r22
    T[0, 3] = xyz[0] + p0
    T[1, 3] = xyz[1] + p1
    T[2, 3] = xyz[2] + p2
    T[3, 3] = 1.0
    return T


def revolute(xyz, rpy, axis, qi):
    T = cs.SX.zeros(4, 4)

    # Origin rotation from RPY ZYX convention
    cr = cs.cos(rpy[0])
    sr = cs.sin(rpy[0])
    cp = cs.cos(rpy[1])
    sp = cs.sin(rpy[1])
    cy = cs.cos(rpy[2])
    sy = cs.sin(rpy[2])
    r00 = cy*cp
    r01 = cy*sp*sr - sy*cr
    r02 = cy*sp*cr + sy*sr
    r10 = sy*cp
    r11 = sy*sp*sr + cy*cr
    r12 = sy*sp*cr - cy*sr
    r20 = -sp
    r21 = cp*sr
    r22 = cp*cr

    # joint rotation from skew sym axis angle
    cqi = cs.cos(qi)
    sqi = cs.sin(qi)
    s00 = (1 - cqi)*axis[0]*axis[0] + cqi
    s11 = (1 - cqi)*axis[1]*axis[1] + cqi
    s22 = (1 - cqi)*axis[2]*axis[2] + cqi
    s01 = (1 - cqi)*axis[0]*axis[1] - axis[2]*sqi
    s10 = (1 - cqi)*axis[0]*axis[1] + axis[2]*sqi
    s12 = (1 - cqi)*axis[1]*axis[2] - axis[0]*sqi
    s21 = (1 - cqi)*axis[1]*axis[2] + axis[0]*sqi
    s20 = (1 - cqi)*axis[0]*axis[2] - axis[1]*sqi
    s02 = (1 - cqi)*axis[0]*axis[2] + axis[1]*sqi

    # Homogeneous transformation matrix
    T[0, 0] = r00*s00 + r01*s10 + r02*s20
    T[1, 0] = r10*s00 + r11*s10 + r12*s20
    T[2, 0] = r20*s00 + r21*s10 + r22*s20

    T[0, 1] = r00*s01 + r01*s11 + r02*s21
    T[1, 1] = r10*s01 + r11*s11 + r12*s21
    T[2, 1] = r20*s01 + r21*s11 + r22*s21

    T[0, 2] = r00*s02 + r01*s12 + r02*s22
    T[1, 2] = r10*s02 + r11*s12 + r12*s22
    T[2, 2] = r20*s02 + r21*s12 + r22*s22

    T[0, 3] = xyz[0]
    T[1, 3] = xyz[1]
    T[2, 3] = xyz[2]
    T[3, 3] = 1.0
    return T


def full_symbolic(xyz, rpy):
    """Gives a symbolic transformation matrix."""
    T = cs.SX.zeros(4, 4)
    cr = cs.cos(rpy[0])
    sr = cs.sin(rpy[0])
    cp = cs.cos(rpy[1])
    sp = cs.sin(rpy[1])
    cy = cs.cos(rpy[2])
    sy = cs.sin(rpy[2])
    T[0, 0] = cy*cp
    T[0, 1] = cy*sp*sr - sy*cr
    T[0, 2] = cy*sp*cr + sy*sr
    T[1, 0] = sy*cp
    T[1, 1] = sy*sp*sr + cy*cr
    T[1, 2] = sy*sp*cr - cy*sr
    T[2, 0] = -sp
    T[2, 1] = cp*sr
    T[2, 2] = cp*cr
    T[0, 3] = xyz[0]
    T[1, 3] = xyz[1]
    T[2, 3] = xyz[2]
    T[3, 3] = 1.0
    return T


def numpy_normalize(v):
    nv = np.linalg.norm(v)
    if nv > 0.0:
        v[0] = v[0]/nv
        v[1] = v[1]/nv
        v[2] = v[2]/nv
    return v


def numpy_skew_symmetric(v):
    """Returns a skew symmetric matrix from vector. p q r"""
    return np.array([[0, -v[2], v[1]],
                     [v[2], 0, -v[0]],
                     [-v[1], v[0], 0]])


def numpy_rotation_rpy(roll, pitch, yaw):
    """Returns a rotation matrix from roll pitch yaw. ZYX convention."""
    cr = np.cos(roll)
    sr = np.sin(roll)
    cp = np.cos(pitch)
    sp = np.sin(pitch)
    cy = np.cos(yaw)
    sy = np.sin(yaw)
    return np.array([[cy*cp,  cy*sp*sr - sy*cr,  cy*sp*cr + sy*sr],
                     [sy*cp,  sy*sp*sr + cy*cr,  sy*sp*cr - cy*sr],
                     [  -sp,             cp*sr,             cp*cr]])


def numpy_rpy(displacement, roll, pitch, yaw):
    """Homogeneous transformation matrix with roll pitch yaw."""
    T = np.zeros([4, 4])
    T[:3, :3] = numpy_rotation_rpy(roll, pitch, yaw)
    T[:3, 3] = displacement
    T[3, 3] = 1.0
    return T


def numpy_rotation_distance_from_identity(R1, R2):
    """Rotation matrix distance based on distance from identity matrix.
    See comparisons at: https://link.springer.com/content/pdf/10.1007%2Fs10851-009-0161-2.pdf"""
    return np.linalg.norm(np.eye(1) - np.dot(R1, R2.T))



import numpy as np
import casadi as ca
from forwardkinematics.fksCommon.fk import ForwardKinematics


class PointFk(ForwardKinematics):

    def __init__(self, n):
        super().__init__()
        self._n = 2

    def fk(self, q, i, positionOnly=False, endlink=0.0):
        assert i <= self._n
        if isinstance(q, ca.SX):
            assert(q.shape[0] == self._n)
            return self.casadi(q, i, positionOnly=positionOnly, endlink=endlink)
        elif isinstance(q, np.ndarray):
            assert(q.size == self._n)
            return self.numpy(q, i, positionOnly=positionOnly, endlink=endlink)

    def casadi(self, q, i, positionOnly=False, endlink=0.0):
        fk = ca.SX(np.array([0.0, 0.0]))
        if i > 0:
            fk += q
        return fk

    def numpy(self, q, i, positionOnly=False, endlink=0.0):
        fk = np.array([0.0, 0.0])
        if i > 0:
            fk += q
        return fk


if __name__ == "__main__":
    q_ca = ca.SX.sym("q", 2)
    fkPlanar = PointFk(2)
    q_np = np.array([3.0, 0.0])
    fkCasadi = fkPlanar.fk(q_ca, 1, positionOnly=True)
    print(fkCasadi)
    fkNumpy = fkPlanar.fk(q_np, 1, positionOnly=True)
    print(fkNumpy)

